package core

import (
	"context"
	"crypto/ecdsa"
	"errors"
	"fmt"
	"math/big"
	"time"
	"validator/contract"

	"github.com/btcsuite/btcd/btcec/v2"
	"github.com/ethereum/go-ethereum"
	"github.com/ethereum/go-ethereum/accounts/abi/bind"
	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/common/hexutil"
	"github.com/ethereum/go-ethereum/core/types"
	"github.com/ethereum/go-ethereum/ethclient"
	log "github.com/sirupsen/logrus"
)

type ChainRPC struct {
	rpc               *ethclient.Client
	chainID           *big.Int
	privateKey        *ecdsa.PrivateKey
	storageContract   *contract.AddressStorage
	validatorContract *contract.Validator
}

func newChain(rpc, privateKey, storageCa, validatorCa string) *ChainRPC {
	ethRpc, err := ethclient.Dial(rpc)
	if err != nil {
		panic(err)
	}

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	chainID, err := ethRpc.ChainID(ctx)
	if err != nil {
		panic(err)
	}

	log.WithField("chain_id", chainID.String()).Info("rpc get chain id")

	storage, err := contract.NewAddressStorage(common.HexToAddress(storageCa), ethRpc)
	if err != nil {
		panic(err)
	}

	validator, err := contract.NewValidator(common.HexToAddress(validatorCa), ethRpc)
	if err != nil {
		panic(err)
	}

	binPrivateKey, err := hexutil.Decode(privateKey)
	if err == hexutil.ErrMissingPrefix {
		binPrivateKey, err = hexutil.Decode("0x" + privateKey)
		if err != nil {
			panic(err)
		}
	}
	if err != nil && err != hexutil.ErrMissingPrefix {
		panic(err)
	}

	k, _ := btcec.PrivKeyFromBytes(binPrivateKey)
	return &ChainRPC{rpc: ethRpc, storageContract: storage, validatorContract: validator, chainID: chainID, privateKey: k.ToECDSA()}
}

func (r *ChainRPC) GetContainerAddresses() (addrs []common.Address, err error) {
	return r.storageContract.GetContainerAddresses(nil)
}

func (r *ChainRPC) GetNMAddresses() (addrs []common.Address, err error) {
	return r.storageContract.GetNmAddresses(nil)
}

func (r *ChainRPC) GetWorkloadThreshold(totalWorkload uint64) (threshold *big.Int, err error) {
	return r.validatorContract.GetWorkloadDistribution(nil, big.NewInt(int64(totalWorkload)))
}

// SubmitProofs 调用合约提交root
func (r *ChainRPC) SubmitProofs(dateTimestamp int64, merkleSumTreeRoot, merkleTreeRoot common.Hash) (txHash common.Hash, err error) {
	ctx, cancel := context.WithTimeout(context.Background(), time.Minute)
	defer cancel()
	opts, err := bind.NewKeyedTransactorWithChainID(r.privateKey, r.chainID)
	if err != nil {
		return
	}
	opts.Context = ctx
	for i := 0; i < 10; i++ {
		signedTx, err := r.validatorContract.SubmitMerkleRoot(opts, big.NewInt(dateTimestamp), merkleSumTreeRoot, merkleTreeRoot)
		if err != nil {
			log.WithError(err).Error("submit root, call contract failed, retry after 3 seconds")
			time.Sleep(time.Second * 3)
			continue
		}
		receipt, err := r.WaitForReceipt(ctx, signedTx.Hash())
		if err != nil {
			return common.Hash{}, err
		}
		if receipt.Status != types.ReceiptStatusSuccessful {
			return common.Hash{}, fmt.Errorf("tx %s failed", signedTx.Hash().Hex())
		}
		return signedTx.Hash(), nil
	}
	return common.Hash{}, errors.New("failed to submit proofs, max retry")
}

func (r *ChainRPC) WaitForReceipt(ctx context.Context, txHash common.Hash) (receipt *types.Receipt, err error) {
	for {
		select {
		case <-ctx.Done():
			return nil, ctx.Err()
		default:
			receipt, err = r.rpc.TransactionReceipt(ctx, txHash)
			if err == nil {
				return receipt, nil
			}
			if err == ethereum.NotFound {
				time.Sleep(time.Second * 1)
				continue
			}
			return nil, err
		}
	}

}
